import json

import pytest


class TestOMEZarrHandlerV04:
    def test_basic_workflow(self, ome_zarr_image_v04_path):
        from ngio.ngff_meta import get_ngff_image_meta_handler
        from ngio.ngff_meta.v04.zarr_utils import NgffImageMeta04

        handler = get_ngff_image_meta_handler(
            store=ome_zarr_image_v04_path, meta_mode="image"
        )

        meta = handler.load_meta()
        handler.write_meta(meta)

        with open("tests/data/meta_v04/base_ome_zarr_image_meta.json") as f:
            base_ome_zarr_meta = json.load(f)

        saved_meta = NgffImageMeta04(**handler.group.attrs).model_dump(
            exclude_none=True
        )
        assert saved_meta == base_ome_zarr_meta

    def test_basic_workflow_with_cache(self, ome_zarr_image_v04_path):
        from ngio.ngff_meta import get_ngff_image_meta_handler
        from ngio.ngff_meta.v04.zarr_utils import NgffImageMeta04

        handler = get_ngff_image_meta_handler(
            store=ome_zarr_image_v04_path, meta_mode="image", cache=True
        )

        meta = handler.load_meta()
        handler.write_meta(meta)

        with open("tests/data/meta_v04/base_ome_zarr_image_meta.json") as f:
            base_ome_zarr_meta = json.load(f)

        saved_meta = NgffImageMeta04(**handler.group.attrs).model_dump(
            exclude_none=True
        )
        assert saved_meta == base_ome_zarr_meta

    def test_wrong_axis_order(self):
        from pydantic import ValidationError

        from ngio.ngff_meta.v04.specs import NgffImageMeta04

        with open(
            "tests/data/meta_v04/base_ome_zarr_image_meta_wrong_axis_order.json"
        ) as f:
            base_ome_zarr_meta = json.load(f)

        with pytest.raises(ValidationError):
            NgffImageMeta04(**base_ome_zarr_meta)



"""Implementation of the ROI Table class.

This class follows the roi_table specification at:
https://fractal-analytics-platform.github.io/fractal-tasks-core/tables/
"""

from collections.abc import Iterable
from functools import partial
from typing import Literal

import pandas as pd
import zarr
from pydantic import BaseModel

from ngio.core.roi import WorldCooROI
from ngio.tables._utils import validate_columns
from ngio.tables.v1._generic_table import REQUIRED_COLUMNS, BaseTable, write_table_ad


class ROITableV1Meta(BaseModel):
    """Metadata for the ROI table."""

    fractal_table_version: Literal["1"] = "1"
    type: Literal["roi_table"] = "roi_table"


ORIGIN_COLUMNS = [
    "x_micrometer_original",
    "y_micrometer_original",
]
TRANSLATION_COLUMNS = ["translation_x", "translation_y", "translation_z"]
OPTIONAL_COLUMNS = ORIGIN_COLUMNS + TRANSLATION_COLUMNS


def create_empty_roi_table(
    include_origin: bool = False, include_translation: bool = False
) -> pd.DataFrame:
    """Create an empty ROI table."""
    origin_columns = ORIGIN_COLUMNS if include_origin else []
    translation_columns = TRANSLATION_COLUMNS if include_translation else []

    columns = [
        *REQUIRED_COLUMNS,
        *origin_columns,
        *translation_columns,
    ]
    table = pd.DataFrame(
        index=pd.Index([], name="FieldIndex", dtype="str"), columns=columns
    )

    return table


class ROITableV1:
    """Class to handle fractal ROI tables.

    To know more about the ROI table format, please refer to the
    specification at:
    https://fractal-analytics-platform.github.io/fractal-tasks-core/tables/
    """

    def __init__(
        self,
        group: zarr.Group,
        validate_metadata: bool = True,
        validate_table: bool = True,
        index_key: str | None = None,
    ):
        """Initialize the class from an existing group.

        Args:
            group (zarr.Group): The group containing the
                ROI table.
            validate_metadata (bool): If True, the metadata is validated.
            validate_table (bool): If True, the table is validated.
            index_key (str): The column name to use as the index of the DataFrame.
        """
        if validate_metadata:
            self._meta = ROITableV1Meta(**group.attrs)
        else:
            self._meta = ROITableV1Meta.model_construct(**group.attrs)

        # Validate the table is not implemented for the ROI table
        validators = [
            partial(
                validate_columns,
                required_columns=REQUIRED_COLUMNS,
                optional_columns=OPTIONAL_COLUMNS,
            )
        ]
        validators = validators if validate_table else None

        index_key = "FieldIndex" if index_key is None else index_key

        self._table_handler = BaseTable(
            group=group, index_key=index_key, index_type="str", validators=validators
        )

    @classmethod
    def _new(
        cls,
        parent_group: zarr.Group,
        name: str,
        include_origin: bool = False,
        include_translation: bool = False,
        overwrite: bool = False,
    ):
        """Create a new ROI table.

        Note this method is not meant to be called directly.
        But should be called from the TablesHandler class.

        Args:
            parent_group (zarr.Group): The parent group where the ROI table
                will be created.
            name (str): The name of the ROI table.
            table (DataFrame): The ROI table as a DataFrame.
            include_origin (bool): Whether to include the origin columns in the table.
            include_translation (bool): Whether to include the translation columns
                in the table.
            overwrite (bool): Whether to overwrite the table if it already exists.
        """
        group = parent_group.create_group(name, overwrite=overwrite)

        table = create_empty_roi_table(
            include_origin=include_origin, include_translation=include_translation
        )

        meta = ROITableV1Meta()
        write_table_ad(
            group=group,
            table=table,
            index_key="FieldIndex",
            index_type="str",
            meta=meta,
        )
        return cls(group=group)

    @property
    def meta(self) -> ROITableV1Meta:
        """Return the metadata of the ROI table."""
        return self._meta

    @property
    def table_handler(self) -> BaseTable:
        """Return the table handler."""
        return self._table_handler

    @property
    def table(self) -> pd.DataFrame:
        """Return the ROI table as a DataFrame."""
        return self._table_handler.table

    @table.setter
    def table(self, table: pd.DataFrame):
        """Set the feature table."""
        raise NotImplementedError(
            "Setting the table is not implemented. Please use the 'set_table' method."
        )

    def set_table(self, table: pd.DataFrame):
        """Set the feature table."""
        self._table_handler.set_table(table)

    @property
    def field_indexes(self) -> list[str]:
        """Return a list of all field indexes in the table."""
        return self.table.index.tolist()

    def append_rois(self, rois: Iterable[WorldCooROI] | WorldCooROI) -> None:
        """Append ROIs to the current table.

        Args:
            rois (Iterable[WorldCooROI] | WorldCooROI): The ROIs to append.
        """
        if isinstance(rois, WorldCooROI):
            rois = [rois]

        rois_dict = {}
        for roi in rois:
            field_index = roi.infos.get("FieldIndex", None)
            if field_index is None:
                raise ValueError("Field index is required in the ROI infos.")

            rois_dict[field_index] = {
                "x_micrometer": roi.x,
                "y_micrometer": roi.y,
                "z_micrometer": roi.z,
                "len_x_micrometer": roi.x_length,
                "len_y_micrometer": roi.y_length,
                "len_z_micrometer": roi.z_length,
            }

        table_df = self.table
        new_table_df = pd.DataFrame.from_dict(rois_dict, orient="index")

        if not table_df.empty:
            table_df = pd.concat([table_df, new_table_df], axis=0)
        else:
            table_df = new_table_df

        table_df.index.name = "FieldIndex"
        table_df.index = table_df.index.astype(str)
        self.table = table_df

    def _gater_optional_columns(self, series: pd.Series) -> dict:
        """Gather the optional columns from a series."""
        origin_dict = {}
        for column in OPTIONAL_COLUMNS:
            if column in series.index:
                origin_dict[column] = series[column]

        translation_dict = {}
        for column in TRANSLATION_COLUMNS:
            if column in series.index:
                translation_dict[column] = series[column]
        return {**origin_dict, **translation_dict}

    def get_roi(self, field_index) -> WorldCooROI:
        """Get an ROI from the table."""
        if field_index not in self.field_indexes:
            raise ValueError(f"Field index {field_index} is not in the table")

        table_df = self.table.loc[field_index]
        roi = WorldCooROI(
            x=table_df.loc["x_micrometer"],
            y=table_df.loc["y_micrometer"],
            z=table_df.loc["z_micrometer"],
            x_length=table_df.loc["len_x_micrometer"],
            y_length=table_df.loc["len_y_micrometer"],
            z_length=table_df.loc["len_z_micrometer"],
            unit="micrometer",
            infos={"FieldIndex": field_index, **self._gater_optional_columns(table_df)},
        )
        return roi

    @property
    def rois(self) -> list[WorldCooROI]:
        """List all ROIs in the table."""
        return [self.get_roi(field_index) for field_index in self.field_indexes]

    def write(self) -> None:
        """Write the crrent state of the table to the Zarr file."""
        self._table_handler.write(self.meta)



import pytest


class TestImageLikeHandler:
    def test_ngff_image(self, ome_zarr_image_v04_path):
        import numpy as np

        from ngio.core.image_like_handler import ImageLike

        image_handler = ImageLike(store=ome_zarr_image_v04_path, path="0")

        assert image_handler.path == "0"
        assert image_handler.pixel_size.zyx == (1.0, 0.1625, 0.1625)
        assert image_handler.axes_names == ["c", "z", "y", "x"]
        assert image_handler.space_axes_names == ["z", "y", "x"]
        assert image_handler.dimensions.shape == (3, 10, 256, 256)
        shape = image_handler.dimensions.shape
        assert image_handler.shape == shape
        assert image_handler.dimensions.z == 10
        assert image_handler.is_3d
        assert not image_handler.is_time_series
        assert image_handler.is_multi_channels

        assert image_handler.on_disk_array.shape == shape
        assert image_handler.on_disk_dask_array.shape == shape

        assert image_handler.get_array(c=0).shape == shape[1:]
        assert (
            image_handler.get_array(c=0, preserve_dimensions=True).shape
            == (1,) + shape[1:]
        )

        image_handler.set_array(patch=np.ones((3, 10, 256, 256), dtype=np.uint16))
        assert image_handler.get_array(c=0, t=0, z=0, x=0, y=0) == 1

        image_handler.consolidate(order=0)

        image_handler_1 = ImageLike(store=ome_zarr_image_v04_path, path="1")
        assert image_handler_1.get_array(c=0, t=0, z=0, x=0, y=0) == 1

    @pytest.mark.skip("Not implemented yet")
    def test_ngff_image_from_pixel_size(self, ome_zarr_image_v04_path):
        from ngio.core.image_like_handler import ImageLike
        from ngio.ngff_meta import PixelSize

        image_handler = ImageLike(
            store=ome_zarr_image_v04_path,
            pixel_size=PixelSize(z=1.0, x=1.3, y=1.3),
            strict=False,
        )

        assert image_handler.path == "3", image_handler.pixel_size.zyx



import pytest
import zarr
from conftest import ZARR_PYTHON_V


class TestGroupUtils:
    @property
    def test_attrs(self) -> dict:
        return {"a": 1, "b": 2, "c": 3}

    def test_open_group_wrapper(self, store_fixture):
        from ngio.io import open_group_wrapper

        store, zarr_format = store_fixture
        group = open_group_wrapper(store=store, mode="r+", zarr_format=zarr_format)
        group.attrs.update(self.test_attrs)
        assert dict(group.attrs) == self.test_attrs

    @pytest.mark.skipif(ZARR_PYTHON_V, reason="Zarr V2 does not support remote stores.")
    def test_raise_not_implemented_error(self):
        from ngio.io._zarr_group_utils import open_group_wrapper

        with pytest.raises(NotImplementedError):
            open_group_wrapper(
                store=zarr.store.RemoteStore(url="https://test.com/test.zarr"),
                mode="r",
                zarr_format=3,
            )



"""A module to handle OME-NGFF images stored in Zarr format."""

from typing import Literal

import zarr

from ngio._common_types import ArrayLike
from ngio.core.image_handler import Image
from ngio.core.image_like_handler import ImageLike
from ngio.core.roi import WorldCooROI
from ngio.core.utils import create_empty_ome_zarr_label
from ngio.io import StoreLike, StoreOrGroup
from ngio.ngff_meta.fractal_image_meta import LabelMeta, PixelSize


class Label(ImageLike):
    """A class to handle OME-NGFF images stored in Zarr format.

    This class provides methods to load label data and metadata from
    an OME-Zarr file.
    """

    def __init__(
        self,
        store: StoreOrGroup,
        *,
        path: str | None = None,
        idx: int | None = None,
        pixel_size: PixelSize | None = None,
        highest_resolution: bool = False,
        strict: bool = True,
        cache: bool = True,
        label_group=None,
    ) -> None:
        """Initialize the the Label Object.

        Note: Only one of `path`, `idx`, 'pixel_size' or 'highest_resolution'
        should be provided.

        store (StoreOrGroup): The Zarr store or group containing the image data.
        path (str | None): The path to the level.
        idx (int | None): The index of the level.
        pixel_size (PixelSize | None): The pixel size of the level.
        highest_resolution (bool): Whether to get the highest resolution level.
        strict (bool): Whether to raise an error where a pixel size is not found
            to match the requested "pixel_size".
        cache (bool): Whether to cache the metadata.
        label_group: The group containing the labels.
        """
        super().__init__(
            store,
            path=path,
            idx=idx,
            pixel_size=pixel_size,
            highest_resolution=highest_resolution,
            strict=strict,
            meta_mode="label",
            cache=cache,
            _label_group=label_group,
        )

    @property
    def metadata(self) -> LabelMeta:
        """Return the metadata of the image."""
        return super().metadata

    def get_array_from_roi(
        self,
        roi: WorldCooROI,
        t: int | slice | None = None,
        mode: Literal["numpy"] | Literal["dask"] = "numpy",
        preserve_dimensions: bool = False,
    ) -> ArrayLike:
        """Return the label data from a region of interest (ROI).

        Args:
            roi (WorldCooROI): The region of interest.
            t (int | slice | None): The time index or slice.
            mode (str): The mode to return the data.
            preserve_dimensions (bool): Whether to preserve the dimensions of the data.
        """
        return super().get_array_from_roi(
            roi=roi, t=t, c=None, mode=mode, preserve_dimensions=preserve_dimensions
        )

    def set_array_from_roi(
        self,
        patch: ArrayLike,
        roi: WorldCooROI,
        t: int | slice | None = None,
        preserve_dimensions: bool = False,
    ) -> None:
        """Set the label data from a region of interest (ROI).

        Args:
            roi (WorldCooROI): The region of interest.
            patch (ArrayLike): The patch to set.
            t (int | slice | None): The time index or slice.
            preserve_dimensions (bool): Whether to preserve the dimensions of the data.
        """
        return super().set_array_from_roi(
            patch=patch, roi=roi, t=t, c=None, preserve_dimensions=preserve_dimensions
        )

    def get_array(
        self,
        x: int | slice | None = None,
        y: int | slice | None = None,
        z: int | slice | None = None,
        t: int | slice | None = None,
        mode: Literal["numpy"] | Literal["dask"] = "numpy",
        preserve_dimensions: bool = False,
    ) -> ArrayLike:
        """Return the label data.

        Args:
            x (int | slice | None): The x index or slice.
            y (int | slice | None): The y index or slice.
            z (int | slice | None): The z index or slice.
            t (int | slice | None): The time index or slice.
            mode (str): The mode to return the data.
            preserve_dimensions (bool): Whether to preserve the dimensions of the data.
        """
        return super().get_array(
            x=x,
            y=y,
            z=z,
            t=t,
            c=None,
            mode=mode,
            preserve_dimensions=preserve_dimensions,
        )

    def set_array(
        self,
        patch: ArrayLike,
        x: int | slice | None = None,
        y: int | slice | None = None,
        z: int | slice | None = None,
        t: int | slice | None = None,
        preserve_dimensions: bool = False,
    ) -> None:
        """Set the label data in the zarr array.

        Args:
            patch (ArrayLike): The patch to set.
            x (int | slice | None): The x index or slice.
            y (int | slice | None): The y index or slice.
            z (int | slice | None): The z index or slice.
            t (int | slice | None): The time index or slice.
            preserve_dimensions (bool): Whether to preserve the dimensions of the data.
        """
        return super().set_array(
            patch=patch,
            x=x,
            y=y,
            z=z,
            t=t,
            c=None,
            preserve_dimensions=preserve_dimensions,
        )

    def mask(
        self,
        roi: WorldCooROI,
        t: int | slice | None = None,
        mode: Literal["numpy"] = "numpy",
        preserve_dimensions: bool = False,
    ) -> ArrayLike:
        """Return the image data from a region of interest (ROI).

        Args:
            roi (WorldCooROI): The region of interest.
            t (int | slice | None): The time index or slice.
            c (int | slice | None): The channel index or slice.
            mask_mode (str): Masking mode
            mode (str): The mode to return the data.
            preserve_dimensions (bool): Whether to preserve the dimensions of the data.
        """
        mask = self.get_array_from_roi(
            roi=roi, t=t, mode=mode, preserve_dimensions=preserve_dimensions
        )

        label = roi.infos.get("label", None)
        if label is None:
            raise ValueError(
                "Label not found in the ROI. Please provide a valid ROI Object."
            )
        mask = mask == label
        return mask

    def consolidate(self) -> None:
        """Consolidate the label group.

        This method consolidates the label group by
        filling all other pyramid levels with the data
        """
        return super().consolidate(order=0)


class LabelGroup:
    """A class to handle the /labels group in an OME-NGFF file."""

    def __init__(
        self,
        group: StoreLike | zarr.Group,
        image_ref: Image | None = None,
        cache: bool = True,
    ) -> None:
        """Initialize the LabelGroupHandler."""
        if not isinstance(group, zarr.Group):
            group = zarr.open_group(group, mode="a")

        if "labels" not in group:
            self._group = group.create_group("labels")
            self._group.attrs["labels"] = []  # initialize the labels attribute
        else:
            self._group: zarr.Group = group["labels"]

        self._image_ref = image_ref
        self._metadata_cache = cache

    def list(self) -> list[str]:
        """List all labels in the group."""
        return self._group.attrs.get("labels", [])

    def num_levels(self, name: str) -> int:
        """Get the number of levels in the labels."""
        label = self.get_label(name)
        return label.metadata.num_levels

    def levels_paths(self, name: str) -> list:
        """Get the paths of the levels in the labels."""
        label = self.get_label(name)
        return label.metadata.levels_paths

    def get_label(
        self,
        name: str,
        path: str | None = None,
        pixel_size: PixelSize | None = None,
        highest_resolution: bool = True,
    ) -> Label:
        """Geta a Label from the group.

        Args:
            name (str): The name of the label.
            path (str | None, optional): The path to the level.
            pixel_size (tuple[float, ...] | list[float] | None, optional): The pixel
                size of the level.
            highest_resolution (bool, optional): Whether to get the highest
                resolution level
        """
        if name not in self.list():
            raise ValueError(f"Label {name} not found in the group.")

        if path is not None or pixel_size is not None:
            highest_resolution = False

        return Label(
            store=self._group[name],
            path=path,
            pixel_size=pixel_size,
            highest_resolution=highest_resolution,
            cache=self._metadata_cache,
        )

    def derive(
        self,
        name: str,
        overwrite: bool = False,
        **kwargs,
    ) -> Label:
        """Derive a new label from an existing label.

        Args:
            name (str): The name of the new label.
            overwrite (bool): If True, the label will be overwritten if it exists.
                Default is False.
            **kwargs: Additional keyword arguments to pass to the new label.
        """
        list_of_labels = self.list()

        if overwrite and name in list_of_labels:
            self._group.attrs["label"] = [
                label for label in list_of_labels if label != name
            ]
        elif not overwrite and name in list_of_labels:
            raise ValueError(f"Label {name} already exists in the group.")

        # create the new label
        new_label_group = self._group.create_group(name, overwrite=overwrite)

        if self._image_ref is None:
            label_0 = self.get_label(list_of_labels[0])
            metadata = label_0.metadata
            on_disk_shape = label_0.on_disk_shape
            chunks = label_0.on_disk_array.chunks
            dataset = label_0.dataset
        else:
            label_0 = self._image_ref
            metadata = label_0.metadata
            channel_index = metadata.index_mapping.get("c", None)
            if channel_index is not None:
                on_disk_shape = (
                    label_0.on_disk_shape[:channel_index]
                    + label_0.on_disk_shape[channel_index + 1 :]
                )
                chunks = (
                    label_0.on_disk_array.chunks[:channel_index]
                    + label_0.on_disk_array.chunks[channel_index + 1 :]
                )
            else:
                on_disk_shape = label_0.on_disk_shape
                chunks = label_0.on_disk_array.chunks

            metadata = metadata.remove_axis("c")
            dataset = metadata.get_highest_resolution_dataset()

        default_kwargs = {
            "store": new_label_group,
            "shape": on_disk_shape,
            "chunks": chunks,
            "dtype": label_0.on_disk_array.dtype,
            "on_disk_axis": dataset.on_disk_axes_names,
            "pixel_sizes": dataset.pixel_size,
            "xy_scaling_factor": metadata.xy_scaling_factor,
            "z_scaling_factor": metadata.z_scaling_factor,
            "time_spacing": dataset.time_spacing,
            "time_units": dataset.time_axis_unit,
            "num_levels": metadata.num_levels,
            "name": name,
            "overwrite": overwrite,
            "version": metadata.version,
        }

        default_kwargs.update(kwargs)

        create_empty_ome_zarr_label(
            **default_kwargs,
        )

        if name not in self.list():
            self._group.attrs["labels"] = [*list_of_labels, name]
        return self.get_label(name)
